"""Provide high-level UDP broadcast for asyncio"""

from asyncio import (
    DatagramProtocol,
    DatagramTransport,
    get_event_loop,
    wait,
    create_task,
    sleep,
    FIRST_COMPLETED
)
from typing import Union, Text
from socket import SOL_SOCKET, SO_BROADCAST
from .client import DatagramClient
from .types import Address

class Broadcast(DatagramClient):
    """High-level UDP broadcaster"""

    def read_queue(self):
        """Reads response queue"""
        values = []
        while not self._queue.empty():
            item = self._queue.get_nowait()
            values.append(item)
        return values

    # pylint: disable-next=arguments-differ
    async def receive(self, timeout=1.0, count=0):
        """
        Wait for an incoming datagram for time (in seconds) and return it.
        This method is a coroutine.
        """
        if self._queue.empty() and self._closed:
            raise IOError("Enpoint is closed")
        exit_conditions = [create_task(sleep(timeout))]
        if count != 0:
            exit_conditions.append(create_task(self._wait_for_count(count)))
        await wait(exit_conditions, return_when=FIRST_COMPLETED)
        if self._closed:
            raise IOError("Enpoint is closed")
        return self.read_queue()

    async def _wait_for_count(self, count: int) -> None:
        while True:
            if self._queue.qsize() == count:
                return
            await sleep(0.15)

class BroadcastProtocol(DatagramProtocol):
    """Datagram broadcast protocol"""
    # pylint: disable=protected-access

    def __init__(self, broadcast: Broadcast):
        self._broadcast = broadcast
        self._transport = None

    def connection_made(self, transport: DatagramTransport):
        self._transport = transport
        sock = transport.get_extra_info("socket")
        sock.setsockopt(SOL_SOCKET, SO_BROADCAST, 1)
        self._broadcast._transport = transport

    def datagram_received(self, data: Union[bytes, Text], addr: Address):
        self._broadcast.feed_datagram(data, addr)

async def open_broadcast(addr: Address) -> Broadcast:
    """Creates datagram broadcast"""
    loop = get_event_loop()
    broadcast = Broadcast()
    await loop.create_datagram_endpoint(
        remote_addr=addr,
        protocol_factory=lambda: BroadcastProtocol(broadcast),
        allow_broadcast=True
    )
    return broadcast



"""Provide high-level UDP endpoints for asyncio"""
from asyncio import (
    DatagramProtocol,
    DatagramTransport,
    get_event_loop
)
from warnings import warn
from random import randint
from .client import DatagramClient
from .types import Address

class Endpoint(DatagramClient):
    """High-level interface for UDP remote enpoints.
    It is initialized with an optional queue size for the incoming datagrams.
    """
    addr: Address = ("", 0)

    # pylint: disable-next=arguments-differ
    def send(self, data):
        """Send a datagram to the remote host."""
        super().send(data, self.addr)

    async def receive(self):
        """
        Wait for an incoming datagram from the remote host.
        This method is a coroutine.
        """
        data, _ = await super().receive()
        return data

class DatagramEndpointProtocol(DatagramProtocol):
    """Datagram protocol for the endpoint high-level interface."""
    # pylint: disable=protected-access

    def __init__(self, endpoint: Endpoint):
        self._endpoint = endpoint

    def connection_made(self, transport: DatagramTransport):
        self._endpoint._transport = transport

    def connection_lost(self, exc: Exception):
        assert exc is None
        if self._endpoint._write_ready_future is not None:
            self._endpoint._write_ready_future.set_result(None)
        self._endpoint.close()

    def datagram_received(self, data: bytes, addr: str):
        self._endpoint.feed_datagram(data, addr)

    def error_received(self, exc: Exception):
        msg = "Endpoint received an error: {!r}"
        warn(msg.format(exc))

    def pause_writing(self):
        assert self._endpoint._write_ready_future is None
        loop = self._endpoint._transport._loop
        self._endpoint._write_ready_future = loop.create_future()

    def resume_writing(self):
        assert self._endpoint._write_ready_future is not None
        self._endpoint._write_ready_future.set_result(None)
        self._endpoint._write_ready_future = None

async def open_endpoint(address: Address, queue_size=0, **kwargs) -> Endpoint:
    """Open and return a datagram endpoint."""
    loop = get_event_loop()
    endpoint = Endpoint(queue_size)
    endpoint.addr = address
    kwargs['local_addr'] = "0.0.0.0", randint(100, 65535)
    kwargs['protocol_factory'] = lambda: DatagramEndpointProtocol(endpoint)
    await loop.create_datagram_endpoint(**kwargs)
    return endpoint



"""MyrtDesk setup script"""
import setuptools

with open(".version", "r", encoding="utf-8") as fh:
    version = fh.read()

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aiodatagram",
    version=version,
    author="Mikhael Khrustik",
    description="A set of utilities to handle UDP in the asyncio context",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        'aiodatagram',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    python_requires='>=3.7',
    package_dir={'':'.'},
)



"""Datagram client"""

from asyncio import (
    Queue,
    QueueFull
)
from warnings import warn

class DatagramClient:
    """High-level interface for UDP client"""

    def __init__(self, queue_size=0):
        self._queue = Queue(queue_size)
        self._closed = False
        self._transport = None
        self._write_ready_future = None

    def feed_datagram(self, data: bytes, addr: str):
        """Add data to response"""
        try:
            self._queue.put_nowait((data, addr))
        except QueueFull:
            warn('Endpoint queue is full')

    def close(self):
        """Closes datagram"""
        # Manage flag
        if self._closed:
            return
        self._closed = True
        # Wake up
        if self._queue.empty():
            self.feed_datagram(None, None)
        # Close transport
        if self._transport:
            self._transport.close()

    def send(self, data: bytes, addr = None):
        """Send a datagram to the given address."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.sendto(data, addr)

    async def receive(self):
        """Wait for an incoming datagram and return it with
        the corresponding address.
        This method is a coroutine.
        """
        if self._queue.empty() and self._closed:
            raise IOError("Enpoint is closed")
        data, addr = await self._queue.get()
        if data is None:
            raise IOError("Enpoint is closed")
        return data, addr

    def abort(self):
        """Close the transport immediately."""
        if self._closed:
            raise IOError("Enpoint is closed")
        self._transport.abort()
        self.close()

    async def drain(self):
        """Drain the transport buffer below the low-water mark."""
        if self._write_ready_future is not None:
            await self._write_ready_future

    @property
    def address(self):
        """The endpoint address as a (host, port) tuple."""
        return self._transport.get_extra_info("socket").getsockname()

    @property
    def closed(self):
        """Indicates whether the endpoint is closed or not."""
        return self._closed
